@@ -19,9 +19,6 @@
  *
  */
 
-/* CURSORFONT */
-#include <X11/Xlibint.h>
-
 #include "common/xcursor.h"
 #include "common/util.h"
 
@@ -114,7 +111,7 @@ xcursor_font_fromstr(const char *s)
 {
     if(s)
         for(int i = 0; i < countof(xcursor_font); i++)
-            if(xcursor_font[i] && !a_strcmp(s, xcursor_font[i]))
+            if(xcursor_font[i] && A_STREQ(s, xcursor_font[i]))
                 return i;
     return 0;
 }
