@@ -25,18 +25,30 @@
 #define SN_API_NOT_YET_FROZEN
 #include <libsn/sn.h>
 
-#include <ev.h>
+#include <glib.h>
 
 #include <xcb/xcb_icccm.h>
 #include <xcb/xcb_keysyms.h>
 
-#include "key.h"
+#include "objects/key.h"
 #include "color.h"
-#include "font.h"
 #include "common/xembed.h"
 #include "common/buffer.h"
 
-typedef struct wibox_t wibox_t;
+#define ROOT_WINDOW_EVENT_MASK \
+    (const uint32_t []) { \
+        XCB_EVENT_MASK_SUBSTRUCTURE_REDIRECT | XCB_EVENT_MASK_SUBSTRUCTURE_NOTIFY \
+      | XCB_EVENT_MASK_ENTER_WINDOW \
+      | XCB_EVENT_MASK_LEAVE_WINDOW \
+      | XCB_EVENT_MASK_STRUCTURE_NOTIFY \
+      | XCB_EVENT_MASK_BUTTON_PRESS \
+      | XCB_EVENT_MASK_BUTTON_RELEASE \
+      | XCB_EVENT_MASK_FOCUS_CHANGE \
+      | XCB_EVENT_MASK_PROPERTY_CHANGE \
+    }
+
+typedef struct drawable_t drawable_t;
+typedef struct drawin_t drawin_t;
 typedef struct a_screen screen_t;
 typedef struct button_t button_t;
 typedef struct widget_t widget_t;
@@ -47,12 +59,12 @@ ARRAY_TYPE(button_t *, button)
 ARRAY_TYPE(tag_t *, tag)
 ARRAY_TYPE(screen_t, screen)
 ARRAY_TYPE(client_t *, client)
-ARRAY_TYPE(wibox_t *, wibox)
+ARRAY_TYPE(drawin_t *, drawin)
 
 /** Main configuration structure */
 typedef struct
 {
-    /** Xlib Display ref [for Xcursors] */
+    /** Xlib Display */
     Display * display;
     /** Connection ref */
     xcb_connection_t *connection;
@@ -62,8 +74,6 @@ typedef struct
     xcb_key_symbols_t *keysyms;
     /** Logical screens */
     screen_array_t screens;
-    /** True if xinerama is active */
-    bool xinerama_is_active;
     /** Root window key bindings */
     key_array_t keys;
     /** Root window mouse bindings */
@@ -76,68 +86,57 @@ typedef struct
     client_array_t clients;
     /** Embedded windows */
     xembed_window_array_t embedded;
-    /** Path to config file */
-    char *conffile;
     /** Stack client history */
     client_array_t stack;
-    /** Command line passed to awesome */
-    char *argv;
     /** Lua VM state */
     lua_State *L;
     /** All errors messages from loading config files */
     buffer_t startup_errors;
-    /** Default colors */
-    struct
-    {
-        xcolor_t fg, bg;
-    } colors;
-    /** Default font */
-    font_t *font;
-    struct
-    {
-        /** Command to execute when spawning a new client */
-        int manage;
-        /** Command to execute when unmanaging client */
-        int unmanage;
-        /** Command to execute when giving focus to a client */
-        int focus;
-        /** Command to execute when removing focus to a client */
-        int unfocus;
-        /** Command to run when mouse enter a client */
-        int mouse_enter;
-        /** Command to run when mouse leave a client */
-        int mouse_leave;
-        /** Command to run when client list changes */
-        int clients;
-        /** Command to run on numbers of tag changes */
-        int tags;
-        /** Command to run when client gets (un)tagged */
-        int tagged;
-        /** Command to run on property change */
-        int property;
-        /** Command to run on time */
-        int timer;
-        /** Command to run on awesome exit */
-        int exit;
-    } hooks;
-    /** The event loop */
-    struct ev_loop *loop;
-    /** The timeout after which we need to stop select() */
-    struct ev_timer timer;
+    /** main loop that awesome is running on */
+    GMainLoop *loop;
     /** The key grabber function */
     int keygrabber;
     /** The mouse pointer grabber function */
     int mousegrabber;
-    /** Focused screen */
-    screen_t *screen_focus;
-    /** Need to call client_stack_refresh() */
-    bool client_need_stack_refresh;
-    /** Wiboxes */
-    wibox_array_t wiboxes;
+    /** Input focus information */
+    struct
+    {
+        /** Focused client */
+        client_t *client;
+        /** Is there a focus change pending? */
+        bool need_update;
+    } focus;
+    /** Drawins */
+    drawin_array_t drawins;
     /** The startup notification display struct */
     SnDisplay *sndisplay;
     /** Latest timestamp we got from the X server */
     xcb_timestamp_t timestamp;
+    /** Window that contains the systray */
+    struct
+    {
+        xcb_window_t window;
+        /** Systray window parent */
+        drawin_t *parent;
+    } systray;
+    /** The monitor of startup notifications */
+    SnMonitorContext *snmonitor;
+    /** The visual, used to draw */
+    xcb_visualtype_t *visual;
+    /** The screen's default visual */
+    xcb_visualtype_t *default_visual;
+    /** The screen's information */
+    xcb_screen_t *screen;
+    /** A graphic context. */
+    xcb_gcontext_t gc;
+    /** Our default depth */
+    uint8_t default_depth;
+    /** Our default color map */
+    xcb_colormap_t default_cmap;
+    /** Do we have to reban clients? */
+    bool need_lazy_banning;
+    /** Tag list */
+    tag_array_t tags;
 } awesome_t;
 
 extern awesome_t globalconf;
