@@ -21,8 +21,6 @@
 
 #define _GNU_SOURCE
 
-#include <ev.h>
-
 #include <lua.h>
 #include <lauxlib.h>
 #include <lualib.h>
@@ -31,31 +29,36 @@
 
 #include "awesome.h"
 #include "config.h"
-#include "timer.h"
+#include "objects/timer.h"
 #include "awesome-version-internal.h"
 #include "ewmh.h"
 #include "luaa.h"
 #include "spawn.h"
-#include "tag.h"
-#include "client.h"
+#include "objects/tag.h"
+#include "objects/client.h"
+#include "objects/drawin.h"
+#include "objects/drawable.h"
 #include "screen.h"
 #include "event.h"
 #include "selection.h"
-#include "window.h"
+#include "systray.h"
 #include "common/xcursor.h"
 #include "common/buffer.h"
+#include "common/backtrace.h"
 
 #ifdef WITH_DBUS
-extern const struct luaL_reg awesome_dbus_lib[];
+extern const struct luaL_Reg awesome_dbus_lib[];
 #endif
-extern const struct luaL_reg awesome_hooks_lib[];
-extern const struct luaL_reg awesome_keygrabber_lib[];
-extern const struct luaL_reg awesome_mousegrabber_lib[];
-extern const struct luaL_reg awesome_root_lib[];
-extern const struct luaL_reg awesome_mouse_methods[];
-extern const struct luaL_reg awesome_mouse_meta[];
-extern const struct luaL_reg awesome_screen_methods[];
-extern const struct luaL_reg awesome_screen_meta[];
+extern const struct luaL_Reg awesome_keygrabber_lib[];
+extern const struct luaL_Reg awesome_mousegrabber_lib[];
+extern const struct luaL_Reg awesome_root_lib[];
+extern const struct luaL_Reg awesome_mouse_methods[];
+extern const struct luaL_Reg awesome_mouse_meta[];
+extern const struct luaL_Reg awesome_screen_methods[];
+extern const struct luaL_Reg awesome_screen_meta[];
+
+/** Path to config file */
+static char *conffile;
 
 /** Quit awesome.
  * \param L The Lua VM state.
@@ -64,7 +67,7 @@ extern const struct luaL_reg awesome_screen_meta[];
 static int
 luaA_quit(lua_State *L)
 {
-    ev_unloop(globalconf.loop, 1);
+    g_main_loop_quit(globalconf.loop);
     return 0;
 }
 
@@ -95,6 +98,24 @@ luaA_restart(lua_State *L)
     return 0;
 }
 
+/** Load an image from a given path.
+ * \param L The Lua VM state.
+ * \return The number of elements pushed on stack.
+ * \luastack
+ * \lparam The command line to execute.
+ */
+static int
+luaA_load_image(lua_State *L)
+{
+    const char *filename = luaL_checkstring(L, 1);
+    cairo_surface_t *surface = draw_load_image(L, filename);
+    if (!surface)
+        return 0;
+    /* lua has to make sure to free the ref or we have a leak */
+    lua_pushlightuserdata(L, surface);
+    return 1;
+}
+
 /** UTF-8 aware string length computing.
  * \param L The Lua VM state.
  * \return The number of elements pushed on stack.
@@ -135,7 +156,7 @@ luaAe_next(lua_State *L)
  * \param idx The index number of elements in stack.
  * \return 1 if more elements to come, 0 otherwise.
  */
-int
+static int
 luaA_next(lua_State *L, int idx)
 {
     if(luaL_getmetafield(L, idx, "__next"))
@@ -237,18 +258,6 @@ luaAe_type(lua_State *L)
     return 1;
 }
 
-/** Modified version of os.execute() which use spawn code to reset signal
- * handlers correctly before exec()'ing the new program.
- * \param L The Lua VM state.
- * \return The number of arguments pushed on stack.
- */
-static int
-luaAe_os_execute(lua_State *L)
-{
-    lua_pushinteger(L, spawn_system(luaL_optstring(L, 1, NULL)));
-    return 1;
-}
-
 /** Replace various standards Lua functions with our own.
  * \param L The Lua VM state.
  */
@@ -260,179 +269,23 @@ luaA_fixups(lua_State *L)
     lua_pushcfunction(L, luaA_mbstrlen);
     lua_setfield(L, -2, "wlen");
     lua_pop(L, 1);
-    /* replace os.execute */
-    lua_getglobal(L, "os");
-    lua_pushcfunction(L, luaAe_os_execute);
-    lua_setfield(L, -2, "execute");
-    lua_pop(L, 1);
     /* replace next */
-    lua_pushliteral(L, "next");
     lua_pushcfunction(L, luaAe_next);
-    lua_settable(L, LUA_GLOBALSINDEX);
+    lua_setglobal(L, "next");
     /* replace pairs */
-    lua_pushliteral(L, "pairs");
     lua_pushcfunction(L, luaAe_next);
     lua_pushcclosure(L, luaAe_pairs, 1); /* pairs get next as upvalue */
-    lua_settable(L, LUA_GLOBALSINDEX);
+    lua_setglobal(L, "pairs");
     /* replace ipairs */
-    lua_pushliteral(L, "ipairs");
     lua_pushcfunction(L, luaA_ipairs_aux);
     lua_pushcclosure(L, luaAe_ipairs, 1);
-    lua_settable(L, LUA_GLOBALSINDEX);
+    lua_setglobal(L, "ipairs");
     /* replace type */
-    lua_pushliteral(L, "type");
     lua_pushcfunction(L, luaAe_type);
-    lua_settable(L, LUA_GLOBALSINDEX);
+    lua_setglobal(L, "type");
     /* set selection */
-    lua_pushliteral(L, "selection");
     lua_pushcfunction(L, luaA_selection_get);
-    lua_settable(L, LUA_GLOBALSINDEX);
-}
-
-/** __next function for wtable objects.
- * \param L The Lua VM state.
- * \return The number of elements pushed on stack.
- */
-static int
-luaA_wtable_next(lua_State *L)
-{
-    /* upvalue 1 is content table */
-    if(lua_next(L, lua_upvalueindex(1)))
-        return 2;
-    lua_pushnil(L);
-    return 1;
-}
-
-/** __ipairs function for wtable objects.
- * \param L The Lua VM state.
- * \return The number of elements pushed on stack.
- */
-static int
-luaA_wtable_ipairs(lua_State *L)
-{
-    /* push ipairs_aux */
-    lua_pushvalue(L, lua_upvalueindex(2));
-    /* push content table */
-    lua_pushvalue(L, lua_upvalueindex(1));
-    lua_pushinteger(L, 0);  /* and initial value */
-    return 3;
-}
-
-/** Index function of wtable objects.
- * \param L The Lua VM state.
- * \return The number of elements pushed on stack.
- */
-static int
-luaA_wtable_index(lua_State *L)
-{
-    size_t len;
-    const char *buf;
-
-    lua_pushvalue(L, 2);
-    /* check for size, waiting lua 5.2 and __len on tables */
-    if((buf = lua_tolstring(L, -1, &len)))
-        if(a_tokenize(buf, len) == A_TK_LEN)
-        {
-            lua_pushnumber(L, lua_objlen(L, lua_upvalueindex(1)));
-            return 1;
-        }
-    lua_pop(L, 1);
-
-    /* upvalue 1 is content table */
-    lua_rawget(L, lua_upvalueindex(1));
-    return 1;
-}
-
-/** Newindex function of wtable objects.
- * \param L The Lua VM state.
- * \return The number of elements pushed on stack.
- */
-static int
-luaA_wtable_newindex(lua_State *L)
-{
-    bool invalid = false;
-
-    /* push key on top */
-    lua_pushvalue(L, 2);
-    /* get current key value in content table */
-    lua_rawget(L, lua_upvalueindex(1));
-    /* if value is a widget, notify change */
-    if(lua_istable(L, -1) || luaA_toudata(L, -1, &widget_class))
-        invalid = true;
-
-    lua_pop(L, 1); /* remove value */
-
-    /* if new value is a widget or a table */
-    if(lua_istable(L, 3))
-    {
-        luaA_table2wtable(L);
-        invalid = true;
-    }
-    else if(!invalid && luaA_toudata(L, 3, &widget_class))
-        invalid = true;
-
-    /* upvalue 1 is content table */
-    lua_rawset(L, lua_upvalueindex(1));
-
-    if(invalid)
-        luaA_wibox_invalidate_byitem(L, lua_topointer(L, 1));
-
-    return 0;
-}
-
-/** Convert the top element of the stack to a proxied wtable.
- * \param L The Lua VM state.
- */
-void
-luaA_table2wtable(lua_State *L)
-{
-    if(!lua_istable(L, -1))
-        return;
-
-    lua_newtable(L); /* create *real* content table */
-    lua_createtable(L, 0, 5); /* metatable */
-    lua_pushvalue(L, -2); /* copy content table */
-    lua_pushcfunction(L, luaA_ipairs_aux); /* push ipairs aux */
-    lua_pushcclosure(L, luaA_wtable_ipairs, 2);
-    lua_pushvalue(L, -3); /* copy content table */
-    lua_pushcclosure(L, luaA_wtable_next, 1); /* __next has the content table as upvalue */
-    lua_pushvalue(L, -4); /* copy content table */
-    lua_pushcclosure(L, luaA_wtable_index, 1); /* __index has the content table as upvalue */
-    lua_pushvalue(L, -5); /* copy content table */
-    lua_pushcclosure(L, luaA_wtable_newindex, 1); /* __newindex has the content table as upvalue */
-    /* set metatable field with just pushed closure */
-    lua_setfield(L, -5, "__newindex");
-    lua_setfield(L, -4, "__index");
-    lua_setfield(L, -3, "__next");
-    lua_setfield(L, -2, "__ipairs");
-    /* set metatable impossible to touch */
-    lua_pushliteral(L, "wtable");
-    lua_setfield(L, -2, "__metatable");
-    /* set new metatable on original table */
-    lua_setmetatable(L, -3);
-
-    /* initial key */
-    lua_pushnil(L);
-    /* go through original table */
-    while(lua_next(L, -3))
-    {
-        /* if convert value to wtable */
-        luaA_table2wtable(L);
-        /* copy key */
-        lua_pushvalue(L, -2);
-        /* move key before value */
-        lua_insert(L, -2);
-        /* set same value in content table */
-        lua_rawset(L, -4);
-        /* copy key */
-        lua_pushvalue(L, -1);
-        /* push the new value :-> */
-        lua_pushnil(L);
-        /* set orig[k] = nil */
-        lua_rawset(L, -5);
-    }
-    /* remove content table */
-    lua_pop(L, 1);
+    lua_setglobal(L, "selection");
 }
 
 /** Look for an item: table, function, etc.
@@ -544,85 +397,30 @@ luaA_awesome_index(lua_State *L)
     if(luaA_usemetatable(L, 1, 2))
         return 1;
 
-    size_t len;
-    const char *buf = luaL_checklstring(L, 2, &len);
+    const char *buf = luaL_checkstring(L, 2);
 
-    switch(a_tokenize(buf, len))
+    if(A_STREQ(buf, "conffile"))
     {
-      case A_TK_FONT:
-        {
-            char *font = pango_font_description_to_string(globalconf.font->desc);
-            lua_pushstring(L, font);
-            g_free(font);
-        }
-        break;
-      case A_TK_FONT_HEIGHT:
-        lua_pushnumber(L, globalconf.font->height);
-        break;
-      case A_TK_CONFFILE:
-        lua_pushstring(L, globalconf.conffile);
-        break;
-      case A_TK_FG:
-        luaA_pushxcolor(L, globalconf.colors.fg);
-        break;
-      case A_TK_BG:
-        luaA_pushxcolor(L, globalconf.colors.bg);
-        break;
-      case A_TK_VERSION:
-        lua_pushliteral(L, AWESOME_VERSION);
-        break;
-      case A_TK_RELEASE:
-        lua_pushliteral(L, AWESOME_RELEASE);
-        break;
-      case A_TK_STARTUP_ERRORS:
-        if (globalconf.startup_errors.len == 0)
-            return 0;
-        lua_pushstring(L, globalconf.startup_errors.s);
-      default:
-        return 0;
+        lua_pushstring(L, conffile);
+        return 1;
     }
 
-    return 1;
-}
-
-/** Newindex function for the awesome global table.
- * \param L The Lua VM state.
- * \return The number of elements pushed on stack.
- */
-static int
-luaA_awesome_newindex(lua_State *L)
-{
-    if(luaA_usemetatable(L, 1, 2))
-        return 1;
-
-    size_t len;
-    const char *buf = luaL_checklstring(L, 2, &len);
-
-    switch(a_tokenize(buf, len))
+    if(A_STREQ(buf, "version"))
     {
-      case A_TK_FONT:
-        {
-            const char *newfont = luaL_checkstring(L, 3);
-            draw_font_delete(&globalconf.font);
-            globalconf.font = draw_font_new(newfont);
-            /* refresh all wiboxes */
-            foreach(wibox, globalconf.wiboxes)
-                (*wibox)->need_update = true;
-            foreach(c, globalconf.clients)
-                if((*c)->titlebar)
-                    (*c)->titlebar->need_update = true;
-        }
-        break;
-      case A_TK_FG:
-        if((buf = luaL_checklstring(L, 3, &len)))
-           xcolor_init_reply(xcolor_init_unchecked(&globalconf.colors.fg, buf, len));
-        break;
-      case A_TK_BG:
-        if((buf = luaL_checklstring(L, 3, &len)))
-           xcolor_init_reply(xcolor_init_unchecked(&globalconf.colors.bg, buf, len));
-        break;
-      default:
-        return 0;
+        lua_pushliteral(L, AWESOME_VERSION);
+        return 1;
+    }
+
+    if(A_STREQ(buf, "release"))
+    {
+        lua_pushliteral(L, AWESOME_RELEASE);
+        return 1;
+    }
+
+    if(A_STREQ(buf, "startup_errors") && globalconf.startup_errors.len != 0)
+    {
+        lua_pushstring(L, globalconf.startup_errors.s);
+        return 1;
     }
 
     return 0;
@@ -636,11 +434,11 @@ luaA_awesome_newindex(lua_State *L)
  * \lparam The function to call.
  */
 static int
-luaA_awesome_add_signal(lua_State *L)
+luaA_awesome_connect_signal(lua_State *L)
 {
     const char *name = luaL_checkstring(L, 1);
     luaA_checkfunction(L, 2);
-    signal_add(&global_signals, name, luaA_object_ref(L, 2));
+    signal_connect(&global_signals, name, luaA_object_ref(L, 2));
     return 0;
 }
 
@@ -652,12 +450,12 @@ luaA_awesome_add_signal(lua_State *L)
  * \lparam The function to call.
  */
 static int
-luaA_awesome_remove_signal(lua_State *L)
+luaA_awesome_disconnect_signal(lua_State *L)
 {
     const char *name = luaL_checkstring(L, 1);
     luaA_checkfunction(L, 2);
     const void *func = lua_topointer(L, 2);
-    signal_remove(&global_signals, name, func);
+    signal_disconnect(&global_signals, name, func);
     luaA_object_unref(L, (void *) func);
     return 0;
 }
@@ -679,8 +477,12 @@ luaA_awesome_emit_signal(lua_State *L)
 static int
 luaA_panic(lua_State *L)
 {
-    warn("unprotected error in call to Lua API (%s), restarting awesome",
+    warn("unprotected error in call to Lua API (%s)",
          lua_tostring(L, -1));
+    buffer_t buf;
+    backtrace_get(&buf);
+    warn("dumping backtrace\n%s", buf.s);
+    warn("restarting awesome");
     awesome_restart();
     return 0;
 }
@@ -713,17 +515,18 @@ void
 luaA_init(xdgHandle* xdg)
 {
     lua_State *L;
-    static const struct luaL_reg awesome_lib[] =
+    static const struct luaL_Reg awesome_lib[] =
     {
         { "quit", luaA_quit },
         { "exec", luaA_exec },
         { "spawn", luaA_spawn },
         { "restart", luaA_restart },
-        { "add_signal", luaA_awesome_add_signal },
-        { "remove_signal", luaA_awesome_remove_signal },
+        { "connect_signal", luaA_awesome_connect_signal },
+        { "disconnect_signal", luaA_awesome_disconnect_signal },
         { "emit_signal", luaA_awesome_emit_signal },
+        { "systray", luaA_systray },
+        { "load_image", luaA_load_image },
         { "__index", luaA_awesome_index },
-        { "__newindex", luaA_awesome_newindex },
         { NULL, NULL }
     };
 
@@ -745,26 +548,22 @@ luaA_init(xdgHandle* xdg)
     luaA_openlib(L, "awesome", awesome_lib, awesome_lib);
 
     /* Export root lib */
-    luaL_register(L, "root", awesome_root_lib);
-    lua_pop(L, 1); /* luaL_register() leaves the table on stack */
-
-    /* Export hooks lib */
-    luaL_register(L, "hooks", awesome_hooks_lib);
-    lua_pop(L, 1); /* luaL_register() leaves the table on stack */
+    luaA_registerlib(L, "root", awesome_root_lib);
+    lua_pop(L, 1); /* luaA_registerlib() leaves the table on stack */
 
 #ifdef WITH_DBUS
     /* Export D-Bus lib */
-    luaL_register(L, "dbus", awesome_dbus_lib);
-    lua_pop(L, 1); /* luaL_register() leaves the table on stack */
+    luaA_registerlib(L, "dbus", awesome_dbus_lib);
+    lua_pop(L, 1); /* luaA_registerlib() leaves the table on stack */
 #endif
 
     /* Export keygrabber lib */
-    luaL_register(L, "keygrabber", awesome_keygrabber_lib);
-    lua_pop(L, 1); /* luaL_register() leaves the table on stack */
+    luaA_registerlib(L, "keygrabber", awesome_keygrabber_lib);
+    lua_pop(L, 1); /* luaA_registerlib() leaves the table on stack */
 
     /* Export mousegrabber lib */
-    luaL_register(L, "mousegrabber", awesome_mousegrabber_lib);
-    lua_pop(L, 1); /* luaL_register() leaves the table on stack */
+    luaA_registerlib(L, "mousegrabber", awesome_mousegrabber_lib);
+    lua_pop(L, 1); /* luaA_registerlib() leaves the table on stack */
 
     /* Export screen */
     luaA_openlib(L, "screen", awesome_screen_methods, awesome_screen_meta);
@@ -775,17 +574,17 @@ luaA_init(xdgHandle* xdg)
     /* Export button */
     button_class_setup(L);
 
-    /* Export image */
-    image_class_setup(L);
-
     /* Export tag */
     tag_class_setup(L);
 
-    /* Export wibox */
-    wibox_class_setup(L);
+    /* Export window */
+    window_class_setup(L);
 
-    /* Export widget */
-    widget_class_setup(L);
+    /* Export drawable */
+    drawable_class_setup(L);
+
+    /* Export drawin */
+    drawin_class_setup(L);
 
     /* Export client */
     client_class_setup(L);
@@ -796,20 +595,6 @@ luaA_init(xdgHandle* xdg)
     /* Export timer */
     timer_class_setup(L);
 
-    /* init hooks */
-    globalconf.hooks.manage = LUA_REFNIL;
-    globalconf.hooks.unmanage = LUA_REFNIL;
-    globalconf.hooks.focus = LUA_REFNIL;
-    globalconf.hooks.unfocus = LUA_REFNIL;
-    globalconf.hooks.mouse_enter = LUA_REFNIL;
-    globalconf.hooks.mouse_leave = LUA_REFNIL;
-    globalconf.hooks.clients = LUA_REFNIL;
-    globalconf.hooks.tags = LUA_REFNIL;
-    globalconf.hooks.tagged = LUA_REFNIL;
-    globalconf.hooks.property = LUA_REFNIL;
-    globalconf.hooks.timer = LUA_REFNIL;
-    globalconf.hooks.exit = LUA_REFNIL;
-
     /* add Lua search paths */
     lua_getglobal(L, "package");
     if (LUA_TTABLE != lua_type(L, 1))
@@ -848,6 +633,19 @@ luaA_init(xdgHandle* xdg)
     lua_pushliteral(L, ";" AWESOME_LUA_LIB_PATH "/?/init.lua");
     lua_concat(L, 3); /* concatenate with package.path */
     lua_setfield(L, 1, "path"); /* package.path = "concatenated string" */
+
+    lua_getfield(L, 1, "loaded");
+
+    lua_pop(L, 2); /* pop "package" and "package.loaded" */
+
+    signal_add(&global_signals, "debug::error");
+    signal_add(&global_signals, "debug::deprecation");
+    signal_add(&global_signals, "debug::index::miss");
+    signal_add(&global_signals, "debug::newindex::miss");
+    signal_add(&global_signals, "systray::update");
+    signal_add(&global_signals, "wallpaper_changed");
+    signal_add(&global_signals, "refresh");
+    signal_add(&global_signals, "exit");
 }
 
 static void
@@ -867,14 +665,14 @@ luaA_loadrc(const char *confpath, bool run)
         {
             /* Set the conffile right now so it can be used inside the
              * configuration file. */
-            globalconf.conffile = a_strdup(confpath);
+            conffile = a_strdup(confpath);
             if(lua_pcall(globalconf.L, 0, LUA_MULTRET, 0))
             {
                 const char *err = lua_tostring(globalconf.L, -1);
                 luaA_startup_error(err);
                 fprintf(stderr, "%s\n", err);
                 /* An error happened, so reset this. */
-                globalconf.conffile = NULL;
+                conffile = NULL;
             }
             else
                 return true;
@@ -942,13 +740,6 @@ bailout:
     return ret;
 }
 
-void
-luaA_on_timer(EV_P_ ev_timer *w, int revents)
-{
-    if(globalconf.hooks.timer != LUA_REFNIL)
-        luaA_dofunction_from_registry(globalconf.L, globalconf.hooks.timer, 0, 0);
-}
-
 int
 luaA_class_index_miss_property(lua_State *L, lua_object_t *obj)
 {
@@ -963,4 +754,10 @@ luaA_class_newindex_miss_property(lua_State *L, lua_object_t *obj)
     return 0;
 }
 
+void
+luaA_emit_refresh()
+{
+    signal_object_emit(globalconf.L, &global_signals, "refresh", 0);
+}
+
 // vim: filetype=c:expandtab:shiftwidth=4:tabstop=8:softtabstop=4:textwidth=80
