@@ -22,16 +22,102 @@
 #include <X11/keysym.h>
 #include <X11/XF86keysym.h>
 #include <xcb/xtest.h>
+#include <cairo-xcb.h>
 
 #include "globalconf.h"
-#include "button.h"
-#include "wibox.h"
+#include "objects/button.h"
+#include "objects/drawin.h"
 #include "luaa.h"
-#include "window.h"
+#include "xwindow.h"
 #include "common/xcursor.h"
-#include "common/tokenize.h"
 #include "common/xutil.h"
 
+static void
+root_set_wallpaper_pixmap(xcb_connection_t *c, xcb_pixmap_t p)
+{
+    xcb_get_property_cookie_t prop_c;
+    xcb_get_property_reply_t *prop_r;
+    const xcb_screen_t *screen = globalconf.screen;
+
+    /* We now have the pattern painted to the pixmap p. Now turn p into the root
+     * window's background pixmap.
+     */
+    xcb_change_window_attributes(c, screen->root, XCB_CW_BACK_PIXMAP, &p);
+    xcb_clear_area(c, 0, screen->root, 0, 0, 0, 0);
+
+    prop_c = xcb_get_property_unchecked(c, false,
+            screen->root, ESETROOT_PMAP_ID, XCB_ATOM_PIXMAP, 0, 1);
+
+    /* Theoretically, this should be enough to set the wallpaper. However, to
+     * make pseudo-transparency work, clients need a way to get the wallpaper.
+     * You can't query a window's back pixmap, so properties are (ab)used.
+     */
+    xcb_change_property(c, XCB_PROP_MODE_REPLACE, screen->root, _XROOTPMAP_ID, XCB_ATOM_PIXMAP, 32, 1, &p);
+    xcb_change_property(c, XCB_PROP_MODE_REPLACE, screen->root, ESETROOT_PMAP_ID, XCB_ATOM_PIXMAP, 32, 1, &p);
+
+    /* Now make sure that the old wallpaper is freed (but only do this for ESETROOT_PMAP_ID) */
+    prop_r = xcb_get_property_reply(c, prop_c, NULL);
+    if (prop_r && prop_r->value_len)
+    {
+        xcb_pixmap_t *rootpix = xcb_get_property_value(prop_r);
+        if (rootpix)
+            xcb_kill_client(c, *rootpix);
+    }
+    free(prop_r);
+}
+
+static bool
+root_set_wallpaper(cairo_pattern_t *pattern)
+{
+    xcb_connection_t *c = xcb_connect(NULL, NULL);
+    xcb_pixmap_t p = xcb_generate_id(c);
+    /* globalconf.connection should be connected to the same X11 server, so we
+     * can just use the info from that other connection.
+     */
+    const xcb_screen_t *screen = globalconf.screen;
+    uint16_t width = screen->width_in_pixels;
+    uint16_t height = screen->height_in_pixels;
+    bool result = false;
+    cairo_surface_t *surface;
+    cairo_t *cr;
+
+    if (xcb_connection_has_error(c))
+        goto disconnect;
+
+    /* Create a pixmap and make sure it is already created, because we are going
+     * to use it from the other X11 connection (Juggling with X11 connections
+     * is a really, really bad idea).
+     */
+    xcb_create_pixmap(c, screen->root_depth, p, screen->root, width, height);
+    xcb_aux_sync(c);
+
+    /* Now paint to the picture from the main connection so that cairo sees that
+     * it can tell the X server to copy between the (possible) old pixmap and
+     * the new one directly and doesn't need GetImage and PutImage.
+     */
+    surface = cairo_xcb_surface_create(globalconf.connection, p, draw_default_visual(screen), width, height);
+    cr = cairo_create(surface);
+    /* Paint the pattern to the surface */
+    cairo_set_source(cr, pattern);
+    cairo_set_operator(cr, CAIRO_OPERATOR_SOURCE);
+    cairo_paint(cr);
+    cairo_destroy(cr);
+    cairo_surface_finish(surface);
+    cairo_surface_destroy(surface);
+    xcb_aux_sync(globalconf.connection);
+
+    root_set_wallpaper_pixmap(c, p);
+
+    /* Make sure our pixmap is not destroyed when we disconnect. */
+    xcb_set_close_down_mode(c, XCB_CLOSE_DOWN_RETAIN_PERMANENT);
+
+    result = true;
+disconnect:
+    xcb_flush(c);
+    xcb_disconnect(c);
+    return result;
+}
+
 static xcb_keycode_t
 _string_to_key_code(const char *s)
 {
@@ -60,7 +146,6 @@ _string_to_key_code(const char *s)
  * event, this is a boolean value which if true make the coordinates relatives.
  * \lparam In case of a motion event, this is the X coordinate.
  * \lparam In case of a motion event, this is the Y coordinate.
- * \lparam In case of a motion event, this is the screen number to move on.
  * If not specified, the current one is used.
  */
 static int
@@ -72,59 +157,53 @@ luaA_root_fake_input(lua_State *L)
         return 0;
     }
 
-    size_t tlen;
-    const char *stype = luaL_checklstring(L, 1, &tlen);
+    const char *stype = luaL_checkstring(L, 1);
     uint8_t type, detail;
     int x = 0, y = 0;
-    xcb_window_t root = XCB_NONE;
 
-    switch(a_tokenize(stype, tlen))
+    if (A_STREQ(stype, "key_press"))
     {
-      case A_TK_KEY_PRESS:
         type = XCB_KEY_PRESS;
         if(lua_type(L, 2) == LUA_TSTRING) {
-            detail = _string_to_key_code(lua_tostring(L, 2));
+            detail = _string_to_key_code(lua_tostring(L, 2)); /* keysym */
         } else {
             detail = luaL_checknumber(L, 2); /* keycode */
         }
-        break;
-      case A_TK_KEY_RELEASE:
+    }
+    else if(A_STREQ(stype, "key_release"))
+    {
         type = XCB_KEY_RELEASE;
         if(lua_type(L, 2) == LUA_TSTRING) {
-            detail = _string_to_key_code(lua_tostring(L, 2));
+            detail = _string_to_key_code(lua_tostring(L, 2)); /* keysym */
         } else {
             detail = luaL_checknumber(L, 2); /* keycode */
         }
-        break;
-      case A_TK_BUTTON_PRESS:
+    }
+    else if(A_STREQ(stype, "button_press"))
+    {
         type = XCB_BUTTON_PRESS;
         detail = luaL_checknumber(L, 2); /* button number */
-        break;
-      case A_TK_BUTTON_RELEASE:
+    }
+    else if(A_STREQ(stype, "button_release"))
+    {
         type = XCB_BUTTON_RELEASE;
         detail = luaL_checknumber(L, 2); /* button number */
-        break;
-      case A_TK_MOTION_NOTIFY:
+    }
+    else if(A_STREQ(stype, "motion_notify"))
+    {
         type = XCB_MOTION_NOTIFY;
         detail = luaA_checkboolean(L, 2); /* relative to the current position or not */
         x = luaL_checknumber(L, 3);
         y = luaL_checknumber(L, 4);
-        if(lua_gettop(L) == 5 && !globalconf.xinerama_is_active)
-        {
-            int screen = luaL_checknumber(L, 5) - 1;
-            luaA_checkscreen(screen);
-            root = xutil_screen_get(globalconf.connection, screen)->root;
-        }
-        break;
-      default:
-        return 0;
     }
+    else
+        return 0;
 
     xcb_test_fake_input(globalconf.connection,
                         type,
                         detail,
                         XCB_CURRENT_TIME,
-                        root,
+                        XCB_NONE,
                         x, y,
                         0);
     return 0;
@@ -155,14 +234,9 @@ luaA_root_keys(lua_State *L)
         while(lua_next(L, 1))
             key_array_append(&globalconf.keys, luaA_object_ref_class(L, -1, &key_class));
 
-        int nscreen = xcb_setup_roots_length(xcb_get_setup(globalconf.connection));
-
-        for(int phys_screen = 0; phys_screen < nscreen; phys_screen++)
-        {
-            xcb_screen_t *s = xutil_screen_get(globalconf.connection, phys_screen);
-            xcb_ungrab_key(globalconf.connection, XCB_GRAB_ANY, s->root, XCB_BUTTON_MASK_ANY);
-            window_grabkeys(s->root, &globalconf.keys);
-        }
+        xcb_screen_t *s = globalconf.screen;
+        xcb_ungrab_key(globalconf.connection, XCB_GRAB_ANY, s->root, XCB_BUTTON_MASK_ANY);
+        xwindow_grabkeys(s->root, &globalconf.keys);
 
         return 1;
     }
@@ -231,13 +305,10 @@ luaA_root_cursor(lua_State *L)
     {
         uint32_t change_win_vals[] = { xcursor_new(globalconf.display, cursor_font) };
 
-        for(int screen_nbr = 0;
-            screen_nbr < xcb_setup_roots_length(xcb_get_setup(globalconf.connection));
-            screen_nbr++)
-            xcb_change_window_attributes(globalconf.connection,
-                                         xutil_screen_get(globalconf.connection, screen_nbr)->root,
-                                         XCB_CW_CURSOR,
-                                         change_win_vals);
+        xcb_change_window_attributes(globalconf.connection,
+                                     globalconf.screen->root,
+                                     XCB_CW_CURSOR,
+                                     change_win_vals);
     }
     else
         luaA_warn(L, "invalid cursor %s", cursor_name);
@@ -245,33 +316,106 @@ luaA_root_cursor(lua_State *L)
     return 0;
 }
 
-/** Get the wiboxes attached to a screen.
+/** Get the drawins attached to a screen.
  * \param L The Lua VM state.
  * \return The number of element pushed on stack.
  * \luastack
- * \lreturn A table with all wiboxes.
+ * \lreturn A table with all drawins.
  */
 static int
-luaA_root_wiboxes(lua_State *L)
+luaA_root_drawins(lua_State *L)
 {
-    lua_createtable(L, globalconf.wiboxes.len, 0);
+    lua_createtable(L, globalconf.drawins.len, 0);
 
-    for(int i = 0; i < globalconf.wiboxes.len; i++)
+    for(int i = 0; i < globalconf.drawins.len; i++)
     {
-        luaA_object_push(L, globalconf.wiboxes.tab[i]);
+        luaA_object_push(L, globalconf.drawins.tab[i]);
         lua_rawseti(L, -2, i + 1);
     }
 
     return 1;
 }
 
-const struct luaL_reg awesome_root_lib[] =
+/** Get the screen's wallpaper
+ * \param L The Lua VM state.
+ * \return The number of element pushed on stack.
+ * \luastack
+ * \lreturn A cairo surface for the wallpaper.
+ */
+static int
+luaA_root_wallpaper(lua_State *L)
+{
+    xcb_get_property_cookie_t prop_c;
+    xcb_get_property_reply_t *prop_r;
+    xcb_pixmap_t *rootpix;
+    cairo_surface_t *surface;
+
+    if(lua_gettop(L) == 1)
+    {
+        cairo_pattern_t *pattern = (cairo_pattern_t *)lua_touserdata(L, -1);
+        lua_pushboolean(L, root_set_wallpaper(pattern));
+        /* Don't return the wallpaper, it's too easy to get memleaks */
+        return 1;
+    }
+
+    prop_c = xcb_get_property_unchecked(globalconf.connection, false,
+            globalconf.screen->root, _XROOTPMAP_ID, XCB_ATOM_PIXMAP, 0, 1);
+    prop_r = xcb_get_property_reply(globalconf.connection, prop_c, NULL);
+
+    if (!prop_r || !prop_r->value_len)
+    {
+        free(prop_r);
+        return 0;
+    }
+
+    rootpix = xcb_get_property_value(prop_r);
+    if (!rootpix)
+    {
+        free(prop_r);
+        return 0;
+    }
+
+    /* We can't query the pixmap's values (or even if that pixmap exists at
+     * all), so let's just assume that it uses the default visual and is as
+     * large as the root window. Everything else wouldn't make sense.
+     */
+    surface = cairo_xcb_surface_create(globalconf.connection, *rootpix, globalconf.default_visual,
+            globalconf.screen->width_in_pixels, globalconf.screen->height_in_pixels);
+
+    /* lua has to make sure this surface gets destroyed */
+    lua_pushlightuserdata(L, surface);
+    free(prop_r);
+    return 1;
+}
+
+/** Get the screen's wallpaper
+ * \param L The Lua VM state.
+ * \return The number of element pushed on stack.
+ * \luastack
+ * \lreturn A cairo surface for the wallpaper.
+ */
+static int
+luaA_root_tags(lua_State *L)
+{
+    lua_createtable(L, globalconf.tags.len, 0);
+    for(int i = 0; i < globalconf.tags.len; i++)
+    {
+        luaA_object_push(L, globalconf.tags.tab[i]);
+        lua_rawseti(L, -2, i + 1);
+    }
+
+    return 1;
+}
+
+const struct luaL_Reg awesome_root_lib[] =
 {
     { "buttons", luaA_root_buttons },
     { "keys", luaA_root_keys },
     { "cursor", luaA_root_cursor },
     { "fake_input", luaA_root_fake_input },
-    { "wiboxes", luaA_root_wiboxes },
+    { "drawins", luaA_root_drawins },
+    { "wallpaper", luaA_root_wallpaper },
+    { "tags", luaA_root_tags },
     { NULL, NULL }
 };
 
