@@ -23,16 +23,26 @@
 
 #include <xcb/xcb.h>
 #include <xcb/xinerama.h>
+#include <xcb/randr.h>
 
 #include "screen.h"
 #include "ewmh.h"
-#include "tag.h"
-#include "client.h"
-#include "widget.h"
-#include "wibox.h"
+#include "objects/tag.h"
+#include "objects/client.h"
+#include "objects/drawin.h"
 #include "luaa.h"
 #include "common/xutil.h"
 
+struct screen_output_t
+{
+    /** The XRandR names of the output */
+    char *name;
+    /** The size in millimeters */
+    uint32_t mm_width, mm_height;
+};
+
+ARRAY_FUNCS(screen_output_t, screen_output, DO_NOTHING)
+
 static inline area_t
 screen_xsitoarea(xcb_xinerama_screen_info_t si)
 {
@@ -46,36 +56,122 @@ screen_xsitoarea(xcb_xinerama_screen_info_t si)
     return a;
 }
 
-static xcb_visualtype_t *
-screen_default_visual(xcb_screen_t *s)
+static void
+screen_add(screen_t new_screen)
 {
-    xcb_depth_iterator_t depth_iter = xcb_screen_allowed_depths_iterator(s);
-
-    if(depth_iter.data)
-        for(; depth_iter.rem; xcb_depth_next (&depth_iter))
-            for(xcb_visualtype_iterator_t visual_iter = xcb_depth_visuals_iterator(depth_iter.data);
-                visual_iter.rem; xcb_visualtype_next (&visual_iter))
-                if(s->root_visual == visual_iter.data->visual_id)
-                    return visual_iter.data;
-
-    return NULL;
+    foreach(screen_to_test, globalconf.screens)
+        if(new_screen.geometry.x == screen_to_test->geometry.x
+           && new_screen.geometry.y == screen_to_test->geometry.y)
+            {
+                /* we already have a screen for this area, just check if
+                 * it's not bigger and drop it */
+                screen_to_test->geometry.width =
+                    MAX(new_screen.geometry.width, screen_to_test->geometry.width);
+                screen_to_test->geometry.height =
+                    MAX(new_screen.geometry.height, screen_to_test->geometry.height);
+                return;
+            }
+    signal_add(&new_screen.signals, "property::workarea");
+    screen_array_append(&globalconf.screens, new_screen);
 }
 
-/** Get screens informations and fill global configuration.
- */
-void
-screen_scan(void)
+static bool
+screen_scan_randr(void)
 {
+    /* Check for extension before checking for XRandR */
+    if(xcb_get_extension_data(globalconf.connection, &xcb_randr_id)->present)
+    {
+        xcb_randr_query_version_reply_t *version_reply =
+            xcb_randr_query_version_reply(globalconf.connection,
+                                          xcb_randr_query_version(globalconf.connection, 1, 1), 0);
+        if(version_reply)
+        {
+            p_delete(&version_reply);
+
+            /* A quick XRandR recall:
+             * You have CRTC that manages a part of a SCREEN.
+             * Each CRTC can draw stuff on one or more OUTPUT. */
+            xcb_randr_get_screen_resources_cookie_t screen_res_c = xcb_randr_get_screen_resources(globalconf.connection, globalconf.screen->root);
+            xcb_randr_get_screen_resources_reply_t *screen_res_r = xcb_randr_get_screen_resources_reply(globalconf.connection, screen_res_c, NULL);
+
+            /* Only use the data from XRandR if there is more than one screen
+             * defined. This should work around the broken nvidia driver.  */
+            if (screen_res_r->num_crtcs <= 1)
+            {
+                p_delete(&screen_res_r);
+                return false;
+            }
+
+            /* We go through CRTC, and build a screen for each one. */
+            xcb_randr_crtc_t *randr_crtcs = xcb_randr_get_screen_resources_crtcs(screen_res_r);
+
+            for(int i = 0; i < screen_res_r->num_crtcs; i++)
+            {
+                /* Get info on the output crtc */
+                xcb_randr_get_crtc_info_cookie_t crtc_info_c = xcb_randr_get_crtc_info(globalconf.connection, randr_crtcs[i], XCB_CURRENT_TIME);
+                xcb_randr_get_crtc_info_reply_t *crtc_info_r = xcb_randr_get_crtc_info_reply(globalconf.connection, crtc_info_c, NULL);
+
+                /* If CRTC has no OUTPUT, ignore it */
+                if(!xcb_randr_get_crtc_info_outputs_length(crtc_info_r))
+                    continue;
+
+                /* Prepare the new screen */
+                screen_t new_screen;
+                p_clear(&new_screen, 1);
+                new_screen.geometry.x = crtc_info_r->x;
+                new_screen.geometry.y = crtc_info_r->y;
+                new_screen.geometry.width= crtc_info_r->width;
+                new_screen.geometry.height= crtc_info_r->height;
+
+                xcb_randr_output_t *randr_outputs = xcb_randr_get_crtc_info_outputs(crtc_info_r);
+
+                for(int j = 0; j < xcb_randr_get_crtc_info_outputs_length(crtc_info_r); j++)
+                {
+                    xcb_randr_get_output_info_cookie_t output_info_c = xcb_randr_get_output_info(globalconf.connection, randr_outputs[j], XCB_CURRENT_TIME);
+                    xcb_randr_get_output_info_reply_t *output_info_r = xcb_randr_get_output_info_reply(globalconf.connection, output_info_c, NULL);
+
+                    int len = xcb_randr_get_output_info_name_length(output_info_r);
+                    /* name is not NULL terminated */
+                    char *name = memcpy(p_new(char *, len + 1), xcb_randr_get_output_info_name(output_info_r), len);
+                    name[len] = '\0';
+
+                    screen_output_array_append(&new_screen.outputs,
+                                               (screen_output_t) { .name = name,
+                                                                   .mm_width = output_info_r->mm_width,
+                                                                   .mm_height = output_info_r->mm_height });
+
+                    p_delete(&output_info_r);
+                }
+
+                screen_add(new_screen);
+
+                p_delete(&crtc_info_r);
+            }
+
+            p_delete(&screen_res_r);
+
+            return true;
+        }
+    }
+
+    return false;
+}
+
+static bool
+screen_scan_xinerama(void)
+{
+    bool xinerama_is_active = false;
+
     /* Check for extension before checking for Xinerama */
     if(xcb_get_extension_data(globalconf.connection, &xcb_xinerama_id)->present)
     {
         xcb_xinerama_is_active_reply_t *xia;
         xia = xcb_xinerama_is_active_reply(globalconf.connection, xcb_xinerama_is_active(globalconf.connection), NULL);
-        globalconf.xinerama_is_active = xia->state;
+        xinerama_is_active = xia->state;
         p_delete(&xia);
     }
 
-    if(globalconf.xinerama_is_active)
+    if(xinerama_is_active)
     {
         xcb_xinerama_query_screens_reply_t *xsq;
         xcb_xinerama_screen_info_t *xsi;
@@ -91,52 +187,40 @@ screen_scan(void)
         /* now check if screens overlaps (same x,y): if so, we take only the biggest one */
         for(int screen = 0; screen < xinerama_screen_number; screen++)
         {
-            bool drop = false;
-            foreach(screen_to_test, globalconf.screens)
-                if(xsi[screen].x_org == screen_to_test->geometry.x
-                   && xsi[screen].y_org == screen_to_test->geometry.y)
-                    {
-                        /* we already have a screen for this area, just check if
-                         * it's not bigger and drop it */
-                        drop = true;
-                        int i = screen_array_indexof(&globalconf.screens, screen_to_test);
-                        screen_to_test->geometry.width =
-                            MAX(xsi[screen].width, xsi[i].width);
-                        screen_to_test->geometry.height =
-                            MAX(xsi[screen].height, xsi[i].height);
-                    }
-            if(!drop)
-            {
-                screen_t s;
-                p_clear(&s, 1);
-                s.geometry = screen_xsitoarea(xsi[screen]);
-                screen_array_append(&globalconf.screens, s);
-            }
+            screen_t s;
+            p_clear(&s, 1);
+            s.geometry = screen_xsitoarea(xsi[screen]);
+            screen_add(s);
         }
 
         p_delete(&xsq);
 
-        xcb_screen_t *s = xutil_screen_get(globalconf.connection, globalconf.default_screen);
-        globalconf.screens.tab[0].visual = screen_default_visual(s);
+        return true;
     }
-    else
-        /* One screen only / Zaphod mode */
-        for(int screen = 0;
-            screen < xcb_setup_roots_length(xcb_get_setup(globalconf.connection));
-            screen++)
-        {
-            xcb_screen_t *xcb_screen = xutil_screen_get(globalconf.connection, screen);
-            screen_t s;
-            p_clear(&s, 1);
-            s.geometry.x = 0;
-            s.geometry.y = 0;
-            s.geometry.width = xcb_screen->width_in_pixels;
-            s.geometry.height = xcb_screen->height_in_pixels;
-            s.visual = screen_default_visual(xcb_screen);
-            screen_array_append(&globalconf.screens, s);
-        }
 
-    globalconf.screen_focus = globalconf.screens.tab;
+    return false;
+}
+
+static void screen_scan_x11(void)
+{
+    /* One screen only / Zaphod mode */
+    xcb_screen_t *xcb_screen = globalconf.screen;
+    screen_t s;
+    p_clear(&s, 1);
+    s.geometry.x = 0;
+    s.geometry.y = 0;
+    s.geometry.width = xcb_screen->width_in_pixels;
+    s.geometry.height = xcb_screen->height_in_pixels;
+    screen_add(s);
+}
+
+/** Get screens informations and fill global configuration.
+ */
+void
+screen_scan(void)
+{
+    if(!screen_scan_randr() && !screen_scan_xinerama())
+        screen_scan_x11();
 }
 
 /** Return the Xinerama screen number where the coordinates belongs to.
@@ -146,18 +230,15 @@ screen_scan(void)
  * \return Screen pointer or screen param if no match or no multi-head.
  */
 screen_t *
-screen_getbycoord(screen_t *screen, int x, int y)
+screen_getbycoord(int x, int y)
 {
-    /* don't waste our time */
-    if(!globalconf.xinerama_is_active)
-        return screen;
-
     foreach(s, globalconf.screens)
         if((x < 0 || (x >= s->geometry.x && x < s->geometry.x + s->geometry.width))
            && (y < 0 || (y >= s->geometry.y && y < s->geometry.y + s->geometry.height)))
             return s;
 
-    return screen;
+    /* No screen found, let's be creative. */
+    return &globalconf.screens.tab[0];
 }
 
 /** Get screens info.
@@ -207,13 +288,17 @@ screen_area_get(screen_t *screen, bool strut)
     }
 
     foreach(c, globalconf.clients)
-        if(client_isvisible(*c, screen))
+        if((*c)->screen == screen && client_isvisible(*c))
             COMPUTE_STRUT(*c)
 
-    foreach(wibox, globalconf.wiboxes)
-        if((*wibox)->visible
-           && (*wibox)->screen == screen)
-            COMPUTE_STRUT(*wibox)
+    foreach(drawin, globalconf.drawins)
+        if((*drawin)->visible)
+        {
+            screen_t *d_screen =
+                screen_getbycoord((*drawin)->geometry.x, (*drawin)->geometry.y);
+            if (d_screen == screen)
+                COMPUTE_STRUT(*drawin)
+        }
 
 #undef COMPUTE_STRUT
 
@@ -226,13 +311,12 @@ screen_area_get(screen_t *screen, bool strut)
 }
 
 /** Get display info.
- * \param phys_screen Physical screen number.
  * \return The display area.
  */
 area_t
-display_area_get(int phys_screen)
+display_area_get(void)
 {
-    xcb_screen_t *s = xutil_screen_get(globalconf.connection, phys_screen);
+    xcb_screen_t *s = globalconf.screen;
     area_t area = { .x = 0,
                     .y = 0,
                     .width = s->width_in_pixels,
@@ -240,19 +324,6 @@ display_area_get(int phys_screen)
     return area;
 }
 
-/** This returns the real X screen number for a logical
- * screen if Xinerama is active.
- * \param screen The logical screen.
- * \return The X screen.
- */
-int
-screen_virttophys(int screen)
-{
-    if(globalconf.xinerama_is_active)
-        return globalconf.default_screen;
-    return screen;
-}
-
 /** Move a client to a virtual screen.
  * \param c The client to move.
  * \param new_screen The destination screen.
@@ -269,22 +340,13 @@ screen_client_moveto(client_t *c, screen_t *new_screen, bool doresize)
     if(new_screen == c->screen)
         return;
 
-    if (globalconf.screens.tab[c->phys_screen].client_focus == c)
+    if (globalconf.focus.client == c)
         had_focus = true;
 
     c->screen = new_screen;
 
-    if(c->titlebar)
-        c->titlebar->screen = new_screen;
-
-    /* If client was on a screen, remove old tags */
-    if(old_screen)
-        foreach(old_tag, old_screen->tags)
-            untag_client(c, *old_tag);
-
     if(!doresize)
     {
-        hook_property(c, "screen");
         luaA_object_push(globalconf.L, c);
         luaA_object_emit_signal(globalconf.L, -1, "property::screen", 0);
         lua_pop(globalconf.L, 1);
@@ -298,48 +360,23 @@ screen_client_moveto(client_t *c, screen_t *new_screen, bool doresize)
 
     area_t new_geometry = c->geometry;
 
-    if(c->fullscreen)
-    {
-        new_geometry = to;
-        area_t new_f_geometry = c->geometries.fullscreen;
+    new_geometry.x = to.x + new_geometry.x - from.x;
+    new_geometry.y = to.y + new_geometry.y - from.y;
 
-        new_f_geometry.x = to.x + new_f_geometry.x - from.x;
-        new_f_geometry.y = to.y + new_f_geometry.y - from.x;
+    /* resize the client if it doesn't fit the new screen */
+    if(new_geometry.width > to.width)
+        new_geometry.width = to.width;
+    if(new_geometry.height > to.height)
+        new_geometry.height = to.height;
 
-        /* resize the client's original geometry if it doesn't fit the screen */
-        if(new_f_geometry.width > to.width)
-            new_f_geometry.width = to.width;
-        if(new_f_geometry.height > to.height)
-            new_f_geometry.height = to.height;
+    /* make sure the client is still on the screen */
+    if(new_geometry.x + new_geometry.width > to.x + to.width)
+        new_geometry.x = to.x + to.width - new_geometry.width;
+    if(new_geometry.y + new_geometry.height > to.y + to.height)
+        new_geometry.y = to.y + to.height - new_geometry.height;
 
-        /* make sure the client is still on the screen */
-        if(new_f_geometry.x + new_f_geometry.width > to.x + to.width)
-            new_f_geometry.x = to.x + to.width - new_f_geometry.width;
-        if(new_f_geometry.y + new_f_geometry.height > to.y + to.height)
-            new_f_geometry.y = to.y + to.height - new_f_geometry.height;
-
-        c->geometries.fullscreen = new_f_geometry;
-    }
-    else
-    {
-        new_geometry.x = to.x + new_geometry.x - from.x;
-        new_geometry.y = to.y + new_geometry.y - from.y;
-
-        /* resize the client if it doesn't fit the new screen */
-        if(new_geometry.width > to.width)
-           new_geometry.width = to.width;
-        if(new_geometry.height > to.height)
-           new_geometry.height = to.height;
-
-        /* make sure the client is still on the screen */
-        if(new_geometry.x + new_geometry.width > to.x + to.width)
-           new_geometry.x = to.x + to.width - new_geometry.width;
-        if(new_geometry.y + new_geometry.height > to.y + to.height)
-           new_geometry.y = to.y + to.height - new_geometry.height;
-    }
     /* move / resize the client */
     client_resize(c, new_geometry, false);
-    hook_property(c, "screen");
     luaA_object_push(globalconf.L, c);
     luaA_object_emit_signal(globalconf.L, -1, "property::screen", 0);
     lua_pop(globalconf.L, 1);
@@ -355,7 +392,8 @@ screen_client_moveto(client_t *c, screen_t *new_screen, bool doresize)
 static int
 luaA_pushscreen(lua_State *L, screen_t *s)
 {
-    lua_pushlightuserdata(L, s);
+    screen_t **ps = lua_newuserdata(L, sizeof(*ps));
+    *ps = s;
     luaL_getmetatable(L, "screen");
     lua_setmetatable(L, -2);
     return 1;
@@ -370,49 +408,19 @@ luaA_pushscreen(lua_State *L, screen_t *s)
 static int
 luaA_screen_module_index(lua_State *L)
 {
+    const char *name;
+
+    if((name = lua_tostring(L, 2)))
+        foreach(screen, globalconf.screens)
+            foreach(output, screen->outputs)
+                if(A_STREQ(output->name, name))
+                    return luaA_pushscreen(L, screen);
+
     int screen = luaL_checknumber(L, 2) - 1;
     luaA_checkscreen(screen);
     return luaA_pushscreen(L, &globalconf.screens.tab[screen]);
 }
 
-/** Get screen tags.
- * \param L The Lua VM state.
- * \return The number of elements pushed on stack.
- * \luastack
- * \lparam None or a table of tags to set to the screen.
- * The table must contains at least one tag.
- * \return A table with all screen tags.
- */
-static int
-luaA_screen_tags(lua_State *L)
-{
-    screen_t *s = luaL_checkudata(L, 1, "screen");
-
-    if(lua_gettop(L) == 2)
-    {
-        luaA_checktable(L, 2);
-
-        /* Detach all tags, but go backward since the array len will change */
-        for(int i = s->tags.len - 1; i >= 0; i--)
-            tag_remove_from_screen(s->tags.tab[i]);
-
-        lua_pushnil(L);
-        while(lua_next(L, 2))
-            tag_append_to_screen(L, -1, s);
-    }
-    else
-    {
-        lua_createtable(L, s->tags.len, 0);
-        for(int i = 0; i < s->tags.len; i++)
-        {
-            luaA_object_push(L, s->tags.tab[i]);
-            lua_rawseti(L, -2, i + 1);
-        }
-    }
-
-    return 1;
-}
-
 /** A screen.
  * \param L The Lua VM state.
  * \return The number of elements pushed on stack.
@@ -423,30 +431,81 @@ luaA_screen_tags(lua_State *L)
 static int
 luaA_screen_index(lua_State *L)
 {
-    size_t len;
     const char *buf;
+    screen_t **ps;
     screen_t *s;
 
-    if(luaA_usemetatable(L, 1, 2))
-        return 1;
-
-    buf = luaL_checklstring(L, 2, &len);
-    s = lua_touserdata(L, 1);
-
-    switch(a_tokenize(buf, len))
+    /* Get metatable of the screen. */
+    lua_getmetatable(L, 1);
+    /* Get the field */
+    lua_pushvalue(L, 2);
+    lua_rawget(L, -2);
+    /* Do we have a field like that? */
+    if(!lua_isnil(L, -1))
     {
-      case A_TK_INDEX:
-        lua_pushinteger(L, screen_array_indexof(&globalconf.screens, s) + 1);
-        break;
-      case A_TK_GEOMETRY:
-        luaA_pusharea(L, s->geometry);
-        break;
-      case A_TK_WORKAREA:
-        luaA_pusharea(L, screen_area_get(s, true));
-        break;
-      default:
-        return 0;
+        /* Yes, so return it! */
+        lua_remove(L, -2);
+        return 1;
     }
+    /* No, so remove everything. */
+    lua_pop(L, 2);
+
+    buf = luaL_checkstring(L, 2);
+    ps = luaL_checkudata(L, 1, "screen");
+    s = *ps;
+
+    if(A_STREQ(buf, "index"))
+    {
+        lua_pushinteger(L, screen_array_indexof(&globalconf.screens, s) + 1);
+        return 1;
+    }
+
+    if(A_STREQ(buf, "geometry"))
+    {
+        luaA_pusharea(L, s->geometry);
+        return 1;
+    }
+
+    if(A_STREQ(buf, "workarea"))
+    {
+        luaA_pusharea(L, screen_area_get(s, true));
+        return 1;
+    }
+
+    if(A_STREQ(buf, "outputs"))
+    {
+        lua_createtable(L, 0, s->outputs.len);
+        foreach(output, s->outputs)
+        {
+            lua_createtable(L, 2, 0);
+
+            lua_pushinteger(L, output->mm_width);
+            lua_setfield(L, -2, "mm_width");
+            lua_pushinteger(L, output->mm_height);
+            lua_setfield(L, -2, "mm_height");
+
+            lua_setfield(L, -2, output->name);
+        }
+        /* The table of tables we created. */
+        return 1;
+    }
+
+    return 0;
+}
+
+/** A screen.
+ * \param L The Lua VM state.
+ * \return The number of elements pushed on stack.
+ */
+static int
+luaA_screen_equal(lua_State *L)
+{
+    screen_t **ps1;
+    screen_t **ps2;
+
+    ps1 = luaL_checkudata(L, 1, "screen");
+    ps2 = luaL_checkudata(L, 2, "screen");
+    lua_pushboolean(L, *ps1 == *ps2);
 
     return 1;
 }
@@ -457,15 +516,33 @@ luaA_screen_index(lua_State *L)
  * \luastack
  * \lvalue A screen.
  * \lparam A signal name.
- * \lparam A function to call when the signal is emitted.
  */
 static int
 luaA_screen_add_signal(lua_State *L)
 {
-    screen_t *s = lua_touserdata(L, 1);
+    screen_t **ps = luaL_checkudata(L, 1, "screen");
+    screen_t *s = *ps;
+    const char *name = luaL_checkstring(L, 2);
+    signal_add(&s->signals, name);
+    return 0;
+}
+
+/** Connect a screen's signal.
+ * \param L The Lua VM state.
+ * \return The number of elements pushed on stack.
+ * \luastack
+ * \lvalue A screen.
+ * \lparam A signal name.
+ * \lparam A function to call when the signal is emitted.
+ */
+static int
+luaA_screen_connect_signal(lua_State *L)
+{
+    screen_t **ps = luaL_checkudata(L, 1, "screen");
+    screen_t *s = *ps;
     const char *name = luaL_checkstring(L, 2);
     luaA_checkfunction(L, 3);
-    signal_add(&s->signals, name, luaA_object_ref(L, 3));
+    signal_connect(&s->signals, name, luaA_object_ref(L, 3));
     return 0;
 }
 
@@ -478,13 +555,14 @@ luaA_screen_add_signal(lua_State *L)
  * \lparam A function to remove
  */
 static int
-luaA_screen_remove_signal(lua_State *L)
+luaA_screen_disconnect_signal(lua_State *L)
 {
-    screen_t *s = lua_touserdata(L, 1);
+    screen_t **ps = luaL_checkudata(L, 1, "screen");
+    screen_t *s = *ps;
     const char *name = luaL_checkstring(L, 2);
     luaA_checkfunction(L, 3);
     const void *ref = lua_topointer(L, 3);
-    signal_remove(&s->signals, name, ref);
+    signal_disconnect(&s->signals, name, ref);
     luaA_object_unref(L, (void *) ref);
     return 0;
 }
@@ -514,7 +592,8 @@ screen_emit_signal(lua_State *L, screen_t *screen, const char *name, int nargs)
 static int
 luaA_screen_emit_signal(lua_State *L)
 {
-    screen_emit_signal(L, lua_touserdata(L, 1), luaL_checkstring(L, 2), lua_gettop(L) - 2);
+    screen_t **ps = luaL_checkudata(L, 1, "screen");
+    screen_emit_signal(L, *ps, luaL_checkstring(L, 2), lua_gettop(L) - 2);
     return 0;
 }
 
@@ -532,20 +611,21 @@ luaA_screen_count(lua_State *L)
     return 1;
 }
 
-const struct luaL_reg awesome_screen_methods[] =
+const struct luaL_Reg awesome_screen_methods[] =
 {
     { "count", luaA_screen_count },
     { "__index", luaA_screen_module_index },
     { NULL, NULL }
 };
 
-const struct luaL_reg awesome_screen_meta[] =
+const struct luaL_Reg awesome_screen_meta[] =
 {
     { "add_signal", luaA_screen_add_signal },
-    { "remove_signal", luaA_screen_remove_signal },
+    { "connect_signal", luaA_screen_connect_signal },
+    { "disconnect_signal", luaA_screen_disconnect_signal },
     { "emit_signal", luaA_screen_emit_signal },
-    { "tags", luaA_screen_tags },
     { "__index", luaA_screen_index },
+    { "__eq", luaA_screen_equal },
     { NULL, NULL }
 };
 
